/*
 *  Copyright 2001-2010 Mikhail Kryshen <mikhail@kryshen.net>
 *
 *  This file is part of Graph Game.
 *
 *  Graph Game is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Graph Game is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Graph Game.  If not, see <http://www.gnu.org/licenses/>.
 */
package kryshen.graphg;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Enumeration;

class GraphCanvas extends Component implements Runnable {
    Graph graph = null;
    
    Edge selectedEdge = null;

    private Player player1 = null;
    private Player player2 = null;

    private Thread timer = null;
    private long time;
    private long startTime;

    private String message = null;

    final Color normalColor = new Color(90, 170, 90);
    final Color shortedColor = new Color(0, 250, 0);
    final Color highlightColor = new Color(0, 255, 0);
    final Color selectColor = new Color(210, 245, 210);
    final Color markColor = Color.RED;
    final Color messageColor = new Color(2, 165, 2);
    final Font messageFont = new Font("Helvetica", Font.BOLD, 48);

    final int border = 15;
    final int nodeRadius = 4;
    final int highlightNodeRadius = 5;

    final Game game;

    GraphCanvas(Game game) {
	this.game = game;
    }

    synchronized void newGame(Player player1, Player player2) {
	stop();

	message = null;

	this.player1 = player1;
	this.player2 = player2;
	//repaint();

 	startTime = System.currentTimeMillis();
	game.setStatus1(player1.name + "'s move");
	game.setStatus2("Moves: 0");
	game.setStatus3("Time: --:--");

	player1.play();

	time = -1000;
	timer = new Thread(this);
	timer.start();
    }    

    synchronized void stop() {
	timer = null;

	if (this.player1 != null)
	    this.player1.stop();

	if (this.player2 != null)
	    this.player2.stop();
    }

    synchronized void nextTurn(Player player) {
	game.setStatus2("Moves: " + player.moves);

        Path path = graph.shortestPath(graph.start.multiNode,
                graph.finish.multiNode);
        int result = 0;

        if (graph.finish.multiNode == graph.start.multiNode) {
            result = Player.SHORT;
        } else if (graph.shortestPath(graph.start.multiNode,
                graph.finish.multiNode) == null) {
            result = Player.CUT;
        }

	if (result > 0) {
	    if (result == player1.player) {
		game.setStatus1(player1.name + " wins!");
		message = "Game over:\n" + player1.name + " wins!";
	    } else if (result == player2.player) {
		game.setStatus1(player2.name + " wins!");
		message = "Game over:\n" + player2.name + " wins!";
	    }
	    //game.setStatus2("Moves: " + player.moves);
	    game.repaintStatus();
	    repaint();
	    timer = null;
	    
	    return;
	}

	try {
	    Thread.sleep(50);
	} catch (Exception e) {}

	Player next;
        if (player == player1) {
            next = player2;
        } else {
            next = player1;
        }
	game.setStatus1(next.name + "'s move");
	game.setStatus2("Moves: " + next.moves);
	next.play();
	game.repaintStatus();
    }

    public void run() {
	while (Thread.currentThread() == timer) {

	    long oldTime = time;
	    time = System.currentTimeMillis() - startTime;

	    if (isVisible()) {
		if (time / 1000 != oldTime / 1000) {
		    int m = (int)(time / 60000);
		    int s = (int)(time / 1000) - m * 60;
		    game.setStatus3("Time: " + ((m < 10) ? "0" : "") + m + 
				((s < 10) ? ":0" : ":") + s);
		    game.repaintStatus();
		}
	    } else {
		startTime += (time - oldTime);
		time = oldTime;
	    }

	    try {
		Thread.sleep(100);
	    } catch (Exception e) {
		System.err.println(e);
	    }
	}

	if (timer == null) {
	    try {
		Thread.sleep(2000);
	    } catch (InterruptedException e) {
		System.err.println(e);
	    }
	
	    synchronized (this) {
		if (timer == null && message != null) {
		    message = null;
		    repaint();
		}
	    }
	}
    }

    void setSelectedEdge(Edge link) {
	if (selectedEdge == link)
	    return;

	if (selectedEdge != null && selectedEdge.equals(link))
	    return;

	Edge oldEdge = selectedEdge;
	selectedEdge = link;
	repaint(oldEdge);
	repaint(link);
    }

    public synchronized void paint(Graphics g) {
	RenderingManager.setAA(g, RenderingManager.AA_FULL);

	Rectangle cb = g.getClipBounds();

	//g.setColor(Color.red);
	//g.drawRect(cb.x, cb.y, cb.width - 1, cb.height - 1);

	int x1, x2, y1, y2;

	//System.out.println(this+".paint("+g+");");
	//g.drawImage(GamePane.imgBackground, -getLocation().x, -getLocation().y, null);

        if (graph == null) {
            return;
        }

        Dimension size = getSize();

	float scaleX = (size.width - border * 2) / (float) graph.width;
	float scaleY = (size.height - border * 2) / (float) graph.height;

        Enumeration iter = graph.edges.elements();
        while (iter.hasMoreElements()) {
            Edge link = (Edge) iter.nextElement();
            if (link == selectedEdge) {
                g.setColor(selectColor);
            } else if (link.marked) {
                g.setColor(markColor);
            } else if (link.shorted) {
                g.setColor(shortedColor);
            } else {
                g.setColor(normalColor);
            }

            x1 = Math.round(link.node1.x * scaleX) + border;
            y1 = Math.round(link.node1.y * scaleY) + border;
            x2 = Math.round(link.node2.x * scaleX) + border;
            y2 = Math.round(link.node2.y * scaleY) + border;

            if (Math.max(x1, x2) >= cb.x - 1 &&
                    Math.min(x1, x2) <= cb.x + cb.width + 1 &&
                    Math.max(y1, y2) >= cb.y - 1 &&
                    Math.min(y1, y2) <= cb.y + cb.height + 1) {

                g.drawLine(x1, y1, x2, y2);

                if (link.shorted) {
                    g.drawLine(x1 - 1, y1 - 1, x2 - 1, y2 - 1);
                    g.drawLine(x1 + 1, y1 - 1, x2 + 1, y2 - 1);
                    g.drawLine(x1 - 1, y1 + 1, x2 - 1, y2 + 1);
                    g.drawLine(x1 + 1, y1 + 1, x2 + 1, y2 + 1);
                }
            }
        }

	iter = graph.nodes.elements();
        while (iter.hasMoreElements()) {
            Node node = (Node) iter.nextElement();

            int r;

            if (node.equals(graph.start) || node.equals(graph.finish)) {
                g.setColor(highlightColor);
                r = highlightNodeRadius;
            } else {
                g.setColor(normalColor);
                r = nodeRadius;
            }

            x1 = Math.round(node.x * scaleX) + border - r;
            y1 = Math.round(node.y * scaleY) + border - r;
            x2 = x1 + r * 2;
            y2 = y1 + r * 2;

            if (Math.max(x1, x2) >= cb.x &&
                    Math.min(x1, x2) <= cb.x + cb.width &&
                    Math.max(y1, y2) >= cb.y &&
                    Math.min(y1, y2) <= cb.y + cb.height) {

                g.fillOval(x1, y1, r * 2, r * 2);
            }
        }

	RenderingManager.setAA(g, RenderingManager.AA_DEFAULT);

	if (message != null) {            
            g.setFont(messageFont);
	    g.setColor(Color.BLACK);
            RenderingManager.drawStringInRect(message, 1, 1,
                    size.width, size.height,
                    RenderingManager.ALIGN_CENTER, g);
            g.setColor(messageColor);
            RenderingManager.drawStringInRect(message, 0, 0,
                    size.width, size.height,
                    RenderingManager.ALIGN_CENTER, g);
	}
    }
    
    synchronized void repaint(Edge link) {
	if (link == null || graph == null)
	    return;
  
	float kx = (getSize().width - border * 2) / (float) graph.width;
	float ky = (getSize().height - border * 2) / (float) graph.height;

	int x1 = Math.round(link.node1.x * kx) + border;
	int y1 = Math.round(link.node1.y * ky) + border;
	int x2 = Math.round(link.node2.x * kx) + border;
	int y2 = Math.round(link.node2.y * ky) + border;

	repaint(Math.min(x1, x2) - 2, Math.min(y1, y2) - 2,
                Math.abs(x2 - x1) + 4, Math.abs(y2 - y1) + 4);
    }
}

