/*
 *  Copyright 2001-2010 Mikhail Kryshen <mikhail@kryshen.net>
 *
 *  This file is part of Graph Game.
 *
 *  Graph Game is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Graph Game is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Graph Game.  If not, see <http://www.gnu.org/licenses/>.
 */
package kryshen.graphg;

import java.awt.Point;
import java.util.Enumeration;

public class GraphGenerator {

    /** Количество вершин в графе */
    private final int nnodes = 60;//120;//60;

    /** 
     * Минимальное расстояние между вершинами;
     * должно соответствовать nnodes, width и height.
     */
    private final int nodeInterval = 10;//5;//10;

    /**
     * Относительные размеры поля (не в пикселах)
     */
    private final int width = 150;
    private final int height = 100;
    
    private final int edgeDistance = 3;
    
    int minEdgesForNode = 3;
    int maxEdgesForNode = 4;

    private int nodes[][];/* = new int[width][height];*/
    private Graph graph = new Graph();

    public GraphGenerator() {
	graph.width = this.width;
	graph.height = this.height;
    }

    private void createNodes() {
	nodes = new int[width][height];
	graph.nodes.removeAllElements();
	int nn = 0;
	graph.start = addNode(rnd(3, 20), rnd(3, 20));
	graph.finish = addNode(rnd(width - 3, width - 20),
			    rnd(height - 3, height - 20));
	while(nn < nnodes - 2) {
	    int ti = rnd(0, width - 1);
	    int tj = rnd(0, height - 1);
	    if( nodes[ti][tj] == 0 ) {
		addNode(ti, tj);
		nn++;
	    }
	}	
    }
    
    private Node addNode(int ti, int tj) {
	int tis = ti - nodeInterval;
	if (tis < 0) tis = 0;
	int tjs = tj - nodeInterval;
	if (tjs < 0) tjs = 0;
	int tie = ti + nodeInterval;
	if (tie > width - 1) tie = width - 1;
	int tje = tj + nodeInterval;
	if (tje > height - 1) tje = height - 1;
	for(int i = tis; i <= tie; i++)
	    for(int j = tjs; j <= tje; j++) {
		nodes[i][j] = -1;
	    }
	nodes[ti][tj] = 1;
	Node gn = new Node(ti, tj, graph);
	graph.addNode(gn);
	return gn;
    }

    private boolean createEdges() {
        graph.edges.removeAllElements();
        Enumeration iter = graph.nodes.elements();
        while (iter.hasMoreElements()) {
            Node node = (Node) iter.nextElement();
            if (createEdgesFor(node) < 2) {
                return false;
            }
        }
        return true;
    }

    private int createEdgesFor(Node node) {
        int nlinks = rnd(minEdgesForNode, maxEdgesForNode);
        int nl = node.edges.size();

        Node a[] = new Node[graph.nodes.size()];
        Enumeration enu = graph.nodes.elements();
        int nts = -1;

        while (enu.hasMoreElements()) {
            Node tgn = (Node) enu.nextElement();
            if (tgn.edges.size() < maxEdgesForNode) {
                nts++;
                a[nts] = tgn;
            }
        }

        quickSortNodes(a, 0, nts, node);

        for (int i = 1; i <= nts; i++) {
            Edge gl = new Edge(node, a[i]);
            if (isValidEdge(gl)) {
                if (graph.addEdge(gl)) {
                    nl++;
                }
                if (nl >= nlinks) {
                    return nl;
                }
            }
        }
        return nl;
    }
    
    private boolean isValidEdge(Edge edge) {
        Enumeration iter = graph.edges.elements();
        while (iter.hasMoreElements()) {
            Edge edge1 = (Edge) iter.nextElement();

            if (!edge.node1.equals(edge1.node1) &&
                    !edge.node1.equals(edge1.node2) &&
                    !edge.node2.equals(edge1.node1) &&
                    !edge.node2.equals(edge1.node2) &&
                    edge.crosses(edge1)) {
                return false;
            }
        }

        iter = graph.nodes.elements();
        while (iter.hasMoreElements()) {
            Node node = (Node) iter.nextElement();
            if (!node.equals(edge.node1) & !node.equals(edge.node2)) {
                Edge edge1 = new Edge(
                        new Node(node.x - edgeDistance, node.y - edgeDistance),
                        new Node(node.x + edgeDistance, node.y - edgeDistance));

                if (edge.crosses(edge1)) {
                    return false;
                }

                edge1 = new Edge(
                        new Node(node.x - edgeDistance, node.y + edgeDistance),
                        new Node(node.x + edgeDistance, node.y + edgeDistance));

                if (edge.crosses(edge1)) {
                    return false;
                }

                edge1 = new Edge(
                        new Node(node.x - edgeDistance, node.y - edgeDistance),
                        new Node(node.x - edgeDistance, node.y + edgeDistance));

                if (edge.crosses(edge1)) {
                    return false;
                }

                edge1 = new Edge(
                        new Node(node.x + edgeDistance, node.y - edgeDistance),
                        new Node(node.x + edgeDistance, node.y + edgeDistance));

                if (edge.crosses(edge1)) {
                    return false;
                }
            }
        }
        return true;
    }
        
    public static int limit(int s, int l, int u) {
	if (s < l)
	    s = l;
	if (s > u)
	    s = u;
	return s;
    }

    public static Point limitPoint(Point p, int width, int height) {
	return new Point(limit(p.x, 0, width), limit(p.y, 0, height));
    }

    public Graph generateGraph() {
        long t = System.currentTimeMillis();
        boolean ok = false;
        System.out.println(this + ":");
        System.out.println("Making new graph:");
        while (!ok) {
            System.out.print("  Creating nodes...");
            createNodes();
            System.out.println("OK");
            System.out.print("  Creating edges...");
            ok = createEdges();
            System.out.println("OK");
            if (!ok) {
                System.out.println("  Bad graph. Retrying...");
            }
        }
        System.out.println("Done (" +
                (System.currentTimeMillis() - t) / 1000D + " s)");
        return graph;
    }
    
    private Graph getGraph() {
        return graph;
    }

    static int rnd(int lowerbound, int upperbound) {
	return (int)((upperbound - lowerbound + 1) * 
		     Math.random() + lowerbound);
    }

   void quickSortNodes(Node a[], int lo0, int hi0, Node node) {
      int lo = lo0;
      int hi = hi0;
      int mid;

      if (hi0 > lo0) {          
          mid = a[(lo0 + hi0) / 2].distTo(node);
          while (lo <= hi) {
              while ((lo < hi0) && (a[lo].distTo(node) < mid))
                  ++lo;
              while ((hi > lo0) && (a[hi].distTo(node) > mid))
                  --hi;
              if (lo <= hi) {
                  swap(a, lo, hi);
                  ++lo;
                  --hi;
              }
          }
          if (lo0 < hi)
              quickSortNodes(a, lo0, hi, node);
          
          if (lo < hi0)
              quickSortNodes(a, lo, hi0, node);
      }
   }

   private void swap(Object a[], int i, int j) {
      Object T;
      T = a[i];
      a[i] = a[j];
      a[j] = T;
   }
}
